/*
 * Copyright (c) 2002 Valentin Nechayev <netch@netch.kiev.ua> All rights
 * reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met: 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer. 2.
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: io3.c,v 1.8 2003/08/31 18:04:29 netch Exp $
 */

#include "io3.h"
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

void io3_readwait(io3 *);
void io3_writewait(io3 *);
int io3_tv_compare(const struct timeval *, const struct timeval *);
void io3_tv_sub3(const struct timeval *, const struct timeval *, struct timeval *);

io3 *
io3_fdopen(int fd, const char *flags)
{
    int se;
    io3 *p;

    p = calloc(1, sizeof(*p));
    if (!p) {
	errno = ENOMEM;
	return NULL;
    }
    p->io3_file = fdopen(fd, flags);
    if (!p->io3_file) {
	se = errno;
	free(p);
	errno = se;
	return NULL;
    }
    return p;
}

void
io3_close(io3 *p)
{
    if (!p)
	return;
    fclose(p->io3_file);
    free(p);
}

void
io3_set_timeout(io3 *p, const struct timeval *tmo)
{
    if (tmo) {
	p->io3_timeout = *tmo;
	p->io3_flag_has_timeout = 1;
    } else
	p->io3_flag_has_timeout = 0;
}

ssize_t
io3_write(io3 *p, const void *dd, size_t sz)
{
    const char *dc = dd;
    ssize_t nput = 0;

    for (;;) {
	if (sz <= 0)
	    break;
	errno = 0;
	if (putc(*dc, p->io3_file) != EOF) {
	    ++dc;
	    --sz;
	    ++nput;
	    continue;
	}
	if (errno == EAGAIN) {
	    clearerr(p->io3_file);
	    io3_writewait(p);
	}
    }
    if (fflush(p->io3_file) == EOF) {
	if (errno == EAGAIN)
	    clearerr(p->io3_file);
    }
    return nput;
}

int
io3_timed_out(io3 *p)
{
    struct timeval tvnow;

    if (!p->io3_flag_has_timeout)
	return 0;
    gettimeofday(&tvnow, NULL);
    return io3_tv_compare(&tvnow, &p->io3_timeout) >= 0;
}

ssize_t
io3_fngets(io3 *p, char *buf, size_t bsize, int *seen_lf)
{
    size_t pos = 0;

    if (bsize <= 0)
	return 0;
    --bsize;			/* Make space for terminating '\0' */
    for (;;) {
	int c;
	FILE *f = p->io3_file;

	if (pos >= bsize)
	    break;
	c = getc(f);
	if (c == EOF) {
	    if (feof(f))
		break;
	    else if (ferror(f) && errno == EAGAIN) {
		clearerr(p->io3_file);
		if (io3_timed_out(p))
		    break;
		io3_readwait(p);
		continue;
	    } else		/* ??? */
		continue;
	}
	buf[pos++] = c;
	if (c == '\n') {
	    if (seen_lf)
		*seen_lf = 1;
	    break;
	}
    }
    buf[pos] = '\0';
    return pos;
}

void
io3_readwait(io3 *p)
{
    struct timeval tvnow, tvsleep;
    fd_set fdr;
    int fd;

    gettimeofday(&tvnow, NULL);
    if (p->io3_flag_has_timeout &&
	io3_tv_compare(&tvnow, &p->io3_timeout) >= 0)
	return;
    io3_tv_sub3(&p->io3_timeout, &tvnow, &tvsleep);
    fd = fileno(p->io3_file);
    FD_ZERO(&fdr);
    FD_SET(fd, &fdr);
    select(fd + 1, &fdr, NULL, NULL, &tvsleep);
}

void
io3_writewait(io3 *p)
{
    struct timeval tvnow, tvsleep;
    fd_set fdw;
    int fd;

    gettimeofday(&tvnow, NULL);
    if (p->io3_flag_has_timeout &&
	io3_tv_compare(&tvnow, &p->io3_timeout) >= 0) {
	return;
    }
    io3_tv_sub3(&p->io3_timeout, &tvnow, &tvsleep);
    fd = fileno(p->io3_file);
    assert(fd < FD_SETSIZE);
    FD_ZERO(&fdw);
    FD_SET(fd, &fdw);
    select(fd + 1, NULL, &fdw, NULL, &tvsleep);
}

int
io3_tv_compare(const struct timeval *tv1, const struct timeval *tv0)
{
    if (tv1->tv_sec > tv0->tv_sec)
	return 1;
    if (tv1->tv_sec < tv0->tv_sec)
	return -1;
    if (tv1->tv_usec > tv0->tv_usec)
	return 1;
    if (tv1->tv_usec < tv0->tv_usec)
	return -1;
    return 0;
}

void
io3_tv_sub3(const struct timeval *n1, const struct timeval *n0, struct timeval *r)
{
    r->tv_sec = n1->tv_sec - n0->tv_sec;
    r->tv_usec = n1->tv_usec - n0->tv_usec;
    while (r->tv_usec >= 1000000) {
	r->tv_usec -= 1000000;
	++r->tv_sec;
    }
    while (r->tv_usec < 0) {
	r->tv_usec += 1000000;
	--r->tv_sec;
    }
}
