/*
 * Copyright (c) 2002 Valentin Nechayev <netch@netch.kiev.ua> All rights
 * reserved. Based on libmilter server example (appropriate copyright is
 * considered).
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met: 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer. 2.
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: main.c,v 1.26 2003/08/19 16:05:55 netch Exp $
 */

/* Filter for verifying address of sender. */

#include <sys/types.h>
#include <sys/file.h>
#include <sys/queue.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <unistd.h>
#include <pthread.h>
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <assert.h>

#include "libmilter/mfapi.h"

#ifndef FALSE
#define FALSE 0
#endif				/* ! FALSE */
#ifndef TRUE
#define TRUE 1
#endif				/* ! TRUE */

#include "verify.h"
#include "myvers.h"
#include "mcom.h"

#define PRIVP ((Priv*) smfi_getpriv(ctx))

static void
usage(void)
{
	printf(
"Usage: verify_sender [options]\n"
"Where options are:\n"
"-r - 5xx reply to \"MAIL FROM:<>\" leads ro reject sender (default: pass)\n"
"-b - use best MXes only (default: use all MXes on order)\n"
"-m <path> - path to external map (see documentaton)\n"
"-p <socket> - milter socket, as defined in sendmail documentation (MANDATORY)\n"
"\n"
"Example:\n"
"verify_sender -m /etc/mail/verify_sender -p inet:3360@127.0.0.1\n"
);
	_exit(EX_USAGE);
}

static void
trim(char * address)
{
	char * p = address;
	char tmpaddr[MAXADDRLEN];

	while (*p == ' ') p++;
	strlcpy(tmpaddr, p, MAXADDRLEN);
	if ((p = strrchr(tmpaddr, ' '))) {
		while (*p == ' ') { *p = 0; p--;}
	}
	strlcpy(address, tmpaddr, MAXADDRLEN);
}

static void
setreply_allparts(SMFICTX *ctx, const struct address_entry *a)
{
    const char *s1;
    const char *s2;

    if (a->ace_cclass != 4 && a->ace_cclass != 5)
	return;
    s1 = a->ace_cclass == 5 ? "550" : "450";
    s2 = a->ace_cclass == 5 ? "5.7.1" : "4.7.1";
    if (!TAILQ_EMPTY(&a->ace_mmsg)) {
	unsigned n;
	struct ml_entry *mle;
	size_t l;
	char *olbuf;
	char *olpos;

	l = 0;
	n = 0;
	TAILQ_FOREACH(mle, &a->ace_mmsg, mle_link) {
	    l += strlen(mle->mle_text) + (n ? 4 : 0);
	    n++;
	}
	olbuf = xcalloc(l + 1, sizeof(*olbuf));
	n = 0;
	olpos = olbuf;
	TAILQ_FOREACH(mle, &a->ace_mmsg, mle_link) {
	    if (n) {
		strcpy(olpos, " // ");	/* CHECKED */
		olpos += 4;
	    }
	    strcpy(olpos, mle->mle_text);	/* CHECKED */
	    olpos += strlen(olpos);
	    n++;
	}
	*olpos = 0;
	/* Satisfy warnings. */
	smfi_setreply(ctx, (char *)s1, (char *)s2, olbuf);
	free(olbuf);
    } else if (a->ace_msg != NULL) {
	/* Satisfy warnings. */
	smfi_setreply(ctx, (char *)s1, (char *)s2, a->ace_msg);
    }
}

static void
Priv_cleanup(Priv* pp, int f_closeconn)
{
	if (pp->env_from) {
		free(pp->env_from);
		pp->env_from = NULL;
	}
	if (pp->ace) {
		address_entry_free(pp->ace);
		pp->ace = NULL;
	}
	if (f_closeconn)
		pp->relay = 0;
}


sfsistat
mlfi_connect( SMFICTX * ctx, char * resolved, _SOCK_ADDR * addr )
{
	Priv* pp;

	if (!(pp = xcalloc(1, sizeof(*pp)))) return SMFIS_TEMPFAIL;

	pp->env_from = NULL;
	pp->relay = 0;

	if (addr && ((struct sockaddr*)addr)->sa_family == AF_INET) {
		struct sockaddr_in* psia = (struct sockaddr_in*) addr;
		pp->relay = ntohl(psia->sin_addr.s_addr);
	}
	/* save the private data */
	smfi_setpriv(ctx, pp);
	/* continue processing */
	return SMFIS_CONTINUE;
}

static sfsistat
mlfi_envfrom(SMFICTX *ctx, char **envfrom)
{
    char address[MAXADDRLEN];
    int rc;
    struct address_entry *ace;
    size_t n0;
    const char *p0;
    const char *ef = envfrom[0];
    Priv* pp;

    pp = PRIVP;
    assert(pp);
    Priv_cleanup(pp, 0);

    /*
     *   :
     * envfrom     ( milter?);
     * envfrom  "<>" -  (  -
     *   );
     * envfrom   
     */

    if (!ef || strlen(ef) == 0 ||
	!strcmp(ef, "<>")) {
	return SMFIS_CONTINUE;
    }
    n0 = strlen(ef);		/* XXX should use strnlen() */
    if (n0 >= MAXADDRLEN - 1) {
	smfi_setreply(ctx, "554", "5.7.0", "FROM address is too long");
	return SMFIS_REJECT;
    }
    p0 = ef;
    if (*p0 == '<') {
	++p0;
	--n0;
    }
    if (n0 > 0 && p0[n0 - 1] == '>') {
	--n0;
    }
    memmove(address, p0, n0);
    address[n0] = 0;
    /* ,   '@', ģ  */
    p0 = strchr(address, '@');
    if (!p0)
	return SMFIS_REJECT;	/* XXX */
    /*
     * XXX trim(address)
     */
    trim(address);

    /*
     * XXX   'localpart@' -  
     */
    p0 = strchr(address, '@');
    if (!*(++p0))
	    return SMFIS_REJECT;	/* XXX */

    /*
     *     .
     *  ,    .
     */

    if (strspn(address,
	       "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
	       "0123456789.-_@+/=") != strlen(address)) {
	return SMFIS_CONTINUE;
    }

    pp->env_from = xstrdup(address);

    rc = verify_address(address, &ace);

    if (rc < 0 || !ace) {
	if (ace) {
	    address_entry_free(ace);
	    ace = NULL;
	}
	smfi_setreply(ctx, "451", "4.0.0", "Temporary failure");
	return SMFIS_TEMPFAIL;
    }
    pp->ace = ace;
    ace = NULL;
    return SMFIS_CONTINUE;
}

static sfsistat
mlfi_envto(SMFICTX *ctx, char **envto)
{
	Priv* pp;
	int rc, cclass;
	struct address_entry *ace;
	const char *p0;
	const char *et = envto[0];
	char address[MAXADDRLEN];
	size_t n0;


	pp = PRIVP;
	assert(pp);

	/*
	 * pp->env_from can be empty when address is of unknown form. In
	 * mlfi_envfrom(), it is accepted without checking.
	 */
	if(!pp->env_from)
		return SMFIS_CONTINUE;

	if (et && *et) {
		n0 = strlen(et);	/* XXX should use strnlen() */
		p0 = et;
		if (*p0 == '<') {
			++p0;
			--n0;
		}
		if (n0 > 0 && p0[n0 - 1] == '>') --n0;
		memmove(address, p0, n0);
		address[n0] = 0;
		rc = check_extmap(address, pp->relay);
	} else rc = 0;	/* XXX */

	/*
	 * 2, 4, 5 - explicitly pass, tempfail or reject message
	 * any other - act as previously defined by verify_address()
	 */

	switch (rc) {
		case 2:
			return SMFIS_CONTINUE;
		case 4:
			smfi_setreply(ctx, "451", "4.0.0", "By Administrator");
			return SMFIS_TEMPFAIL;
		case 5:
			smfi_setreply(ctx, "551", "5.7.1", "By Administrator");
			return SMFIS_REJECT;
		default:
			if ((ace = pp->ace)) {
				if (ace->ace_msg || !TAILQ_EMPTY(&ace->ace_mmsg))
					setreply_allparts(ctx, ace);
				cclass = ace->ace_cclass;
				ace = NULL;
				if (cclass == 2)
					return SMFIS_CONTINUE;
				if (cclass == 4)
					return SMFIS_TEMPFAIL;
				if (cclass == 5)
					return SMFIS_REJECT;
				smfi_setreply(ctx, "451", "4.0.0", "Software failure");
				return SMFIS_TEMPFAIL;
			} else {
				smfi_setreply(ctx, "451", "4.0.0", "Software failure");
				return SMFIS_TEMPFAIL;
			}
	}
}

static sfsistat
mlfi_abort(SMFICTX* ctx)
{
	Priv* pp;

	pp = PRIVP;
	if (pp)	Priv_cleanup(pp, 0);
	return SMFIS_CONTINUE;
}

sfsistat
mlfi_close(SMFICTX *ctx)
{
	Priv *pp;

	pp = PRIVP;
	if (pp) {
		Priv_cleanup(pp, 1);
		free(pp);
	}
	smfi_setpriv(ctx, NULL);
	return SMFIS_CONTINUE;
}


struct smfiDesc smfilter =
{
    "Verify_Sender",		/* filter name */
    SMFI_VERSION,		/* version code -- do not change */
    0,				/* flags */
    mlfi_connect,		/* connection info filter */
    NULL,			/* SMTP HELO command filter */
    mlfi_envfrom,		/* envelope sender filter */
    mlfi_envto,			/* envelope recipient filter */
    NULL,			/* header filter */
    NULL,			/* end of header */
    NULL,			/* body block filter */
    NULL,			/* end of message */
    mlfi_abort,			/* message aborted */
    mlfi_close			/* connection cleanup */
};

int
main(int argc, char *argv[])
{
    int c;
    const char *args = "m:p:rb";
    const char *MapPath = NULL;

    puts("verify_sender " VERIFY_SENDER_VERSION "\n");

    /* Process command line options */

    if (argc == 1)
	    usage();

    while ((c = getopt(argc, argv, args)) != -1) {
	switch (c) {
	case 'm':
	    MapPath = optarg;
	    break;
	case 'p':
	    if (optarg == NULL || *optarg == '\0') {
		(void)fprintf(stderr, "Illegal conn: %s\n",
			      optarg);
		exit(EX_USAGE);
	    }
	    (void)smfi_setconn(optarg);
	    break;
	case 'b':
	    bestmxonly++;
	    break;
	case 'r':
	    notnullrevrej++;
	    break;
	default:
	    usage();
	    /* NOT REACHED */
	    break;
	}
    }

    verify_init(MapPath);

    if (smfi_register(smfilter) == MI_FAILURE) {
	fprintf(stderr, "smfi_register failed\n");
	exit(EX_UNAVAILABLE);
    }

    if (daemon(0, 0)) {
	return 1;
    }
    return smfi_main();
}
