/*-
 * Copyright (c) 1999 Valentin Nechayev <netch@lucky.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$NetchId: nuxres.c,v 1.1 2000/11/04 18:16:21 netch Exp $
 */

#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <netinet/in.h>		/*  -... */

#ifdef WITH_BIND_PORT
#include <bind/arpa/nameser.h>
#include <bind/netdb.h>
#include <bind/resolv.h>
#else
#include <arpa/nameser.h>
#include <netdb.h>
#include <resolv.h>
#endif

#include <assert.h>
#include "nuxres.h"

static unsigned short 
GetShortF(const unsigned char *p)
{
    return ((255 & p[0]) << 8) + (255 & p[1]);
}

static unsigned long 
GetLongF(const unsigned char *p)
{
    return ((255 & p[0]) << 24) +
    ((255 & p[1]) << 16) +
    ((255 & p[2]) << 8) + (255 & p[3]);
}

/*
 * ND_UR_Free() -    ND_UR. NextRec  ,
 *    
 */

void 
ND_UR_Free(ND_UR *UR)
{
    if (!UR)
	return;
    if (UR->DataType == NDDT_SPECIFIC) {
	switch (UR->Type) {
	case T_SOA:
	    if (UR->Data.SOA) {
		if (UR->Data.SOA->PrimaryNS) free(UR->Data.SOA->PrimaryNS);
		if (UR->Data.SOA->ZoneMaster) free(UR->Data.SOA->ZoneMaster);
		free(UR->Data.SOA);
	    }
	    break;
	case T_MX:
	    if (UR->Data.MX) {
		if (UR->Data.MX->Server) free(UR->Data.MX->Server);
		free(UR->Data.MX);
	    }
	    break;
	case T_NS:
	    if (UR->Data.NS) free(UR->Data.NS);
	    break;
	case T_PTR:
	    if (UR->Data.Ptr) free(UR->Data.Ptr);
	    break;
	case T_CNAME:
	    if (UR->Data.Cname) free(UR->Data.Cname);
	    break;
	}
    }
    if (UR->Name) free(UR->Name);
    free(UR);
}

/* ND_Unpacked_Free() -  ND_Unpacked    */

void 
ND_Unpacked_Free(ND_Unpacked *U)
{
    ND_UR *UR, *UR2;

    if (!U)
	return;
    UR = U->Records;
    while (UR != NULL) {
	UR2 = UR->NextRec;
	ND_UR_Free(UR);
	UR = UR2;
    }
    free(U);
}

/*
 * ND_Unpack() -        .
 *  :    ND_Unpacked  , 
 * ; NULL,  
 * :
 */

ND_Unpacked *
ND_Unpack(int *ierp, const unsigned char *Packed, size_t PackedSize)
{
    int i, ier = NDERR_GENERIC;
    const unsigned char *const End = Packed + PackedSize;
    const unsigned char *const Start = Packed + HFIXEDSZ;
    const unsigned char *Pos = Start;
    const unsigned char *SavedPos;
    unsigned short qdcount;
    register unsigned short ancount, nscount, arcount;
    const HEADER *Hdr;
    ND_Unpacked *U = NULL;
    ND_UR *UR = NULL;
    ND_SoaInfo *SI = NULL;
    ND_MxInfo *MI = NULL;

    if (!Packed || PackedSize < HFIXEDSZ || (ptrdiff_t) PackedSize < 0)
	goto CU;
    Hdr = (const HEADER *) Packed;
    qdcount = GetShortF(Packed + 4);
    ancount = GetShortF(Packed + 6);
    nscount = GetShortF(Packed + 8);
    arcount = GetShortF(Packed + 10);
    U = malloc(sizeof(*U));
    if (!U) {
	ier = NDERR_NOMEM;
	goto CU;
    }
    bzero(U, sizeof(*U));
    U->rcode = Hdr->rcode;

    /* Skip query section */

    for (i = 0; i < qdcount; i++) {
	int             r;

	if (Pos >= End || Pos - Packed >= (ptrdiff_t) PackedSize) {
	    ier = NDERR_BADDATA;
	    goto CU;
	}

	/*       -   */

	r = dn_skipname(Pos, End);
	if (r < 0) {
	    ier = NDERR_BADDATA;
	    goto CU;
	}
	Pos += r + QFIXEDSZ;

    }
    /*  ,      */

    assert(ancount == ntohs(Hdr->ancount));
    assert(nscount == ntohs(Hdr->nscount));
    assert(arcount == ntohs(Hdr->arcount));
    for (i = 0;; i++) {
	char NameBuf[1025];	/**** XXX    ? */
	unsigned DLength;
	int jsec, r1;

	if (UR) {
	    ier = NDERR_SOFTWARE;
	    goto CU;
	}
	if (i >= ancount + nscount + arcount)
	    break;
	if (i >= ancount + nscount)
	    jsec = 4;		/* additional section */
	else if (i >= ancount)
	    jsec = 3;		/* authority section */
	else
	    jsec = 2;		/* answer section */
	/* -    - */
	if (Pos >= End) {
	    ier = NDERR_BADDATA;
	    goto CU;
	}
	UR = calloc(1, sizeof(*UR));
	if (!UR) {
	    ier = NDERR_NOMEM;
	    goto CU;
	}
	UR->Name = NULL;
	UR->DataType = 0;
	UR->Section = jsec;

	/*  :   */

	bzero(NameBuf, sizeof NameBuf);
	r1 = dn_expand(Packed, End, Pos, NameBuf, sizeof NameBuf);
	if (r1 < 0) {
	    ier = NDERR_BADDATA;
	    goto CU;
	}
	UR->Name = strdup(NameBuf);
	if (!UR->Name) {
	    ier = NDERR_NOMEM;
	    goto CU;
	}

	/*  :   */

	Pos += r1;
	if (Pos + 3 * INT16SZ + INT32SZ > End) {
	    ier = NDERR_BADDATA;
	    goto CU;
	}
	UR->Type = GetShortF(Pos);
	UR->Class = GetShortF(Pos + INT16SZ);
	UR->Ttl = GetLongF(Pos + 2 * INT16SZ);
	DLength = GetShortF(Pos + 2 * INT16SZ + INT32SZ);
	Pos += 3 * INT16SZ + INT32SZ;
	if (Pos > End || Pos + DLength > End) {
	    ier = NDERR_BADDATA;
	    goto CU;
	}
	SavedPos = Pos;
	/*  ... */
	switch (UR->Type) {
	    /****   */
	case T_A:
	    ; {
		UR->Data.Addr4 = GetLongF(Pos);
		Pos += INT32SZ;
		UR->DataType = NDDT_SPECIFIC;
	    }
	    break;
	case T_NS:
	    r1 = dn_expand(Packed, End, Pos, NameBuf, sizeof NameBuf);
	    if (r1 < 0) {
		ier = NDERR_BADDATA;
		goto CU;
	    }
	    UR->Data.NS = strdup(NameBuf);
	    if (!UR->Data.NS) {
		ier = NDERR_NOMEM;
		goto CU;
	    }
	    UR->DataType = NDDT_SPECIFIC;
	    break;
	case T_MX:
	    /*  precedence: short  .     */
	    ; {
                assert(!MI);

		if (DLength < 3) {
		    ier = NDERR_BADDATA;
		    goto CU;
		}
		MI = calloc(1, sizeof(*MI));
		if (!MI) {
		    ier = NDERR_NOMEM;
		    goto CU;
		}
		MI->Prec = GetShortF(Pos);
		Pos += INT16SZ;
		r1 = dn_expand(Packed, End, Pos, NameBuf, sizeof NameBuf);
		if (r1 < 0) {
		    ier = NDERR_BADDATA;
		    goto CU;
		}
		if (!(MI->Server = strdup(NameBuf))) {
		    ier = NDERR_NOMEM;
		    goto CU;
		}
		Pos += r1;
		UR->Data.MX = MI;
		UR->DataType = NDDT_SPECIFIC;
                MI = NULL;
	    }
	    break;
	case T_PTR:
	    r1 = dn_expand(Packed, End, Pos, NameBuf, sizeof NameBuf);
	    if (r1 < 0) {
		ier = NDERR_BADDATA;
		goto CU;
	    }
	    if ((UR->Data.Ptr = strdup(NameBuf)) == NULL) {
		ier = NDERR_NOMEM;
		goto CU;
	    }
	    UR->DataType = NDDT_SPECIFIC;
	    break;
	case T_CNAME:
	    r1 = dn_expand(Packed, End, Pos, NameBuf, sizeof NameBuf);
	    if (r1 < 0) {
		ier = NDERR_BADDATA;
		goto CU;
	    }
	    if ((UR->Data.Cname = strdup(NameBuf)) == NULL) {
		ier = NDERR_NOMEM;
		goto CU;
	    }
	    UR->DataType = NDDT_SPECIFIC;
	    break;
	    /**** SOA */
	case T_SOA:
	    ; {
		assert(!SI);
		SI = calloc(1, sizeof(*SI));
		if (!SI) {
		    ier = NDERR_NOMEM;
		    goto CU;
		}
		r1 = dn_expand(Packed, End, Pos, NameBuf, sizeof NameBuf);
		if (r1 < 0) {
		    ier = NDERR_BADDATA;
		    goto CU;
		}
		if ((SI->PrimaryNS = strdup(NameBuf)) == NULL) {
		    ier = NDERR_BADDATA;
		    goto CU;
		}
		Pos += r1;
		r1 = dn_expand(Packed, End, Pos, NameBuf, sizeof NameBuf);
		if (r1 < 0) {
		    ier = NDERR_BADDATA;
		    goto CU;
		}
		if ((SI->ZoneMaster = strdup(NameBuf)) == NULL) {
		    ier = NDERR_BADDATA;
		    goto CU;
		}
		Pos += r1;
		if (Pos + 5 * INT32SZ > End) {
		    ier = NDERR_BADDATA;
		    goto CU;
		}
		SI->Serial = GetLongF(Pos);
		Pos += INT32SZ;
		SI->Refresh = GetLongF(Pos);
		Pos += INT32SZ;
		SI->Retry = GetLongF(Pos);
		Pos += INT32SZ;
		SI->Expiry = GetLongF(Pos);
		Pos += INT32SZ;
		SI->Minimum = GetLongF(Pos);
		Pos += INT32SZ;
		UR->DataType = NDDT_SPECIFIC;
		UR->Data.SOA = SI;
		SI = NULL;
	    }
	    break;
	default:
	    /*     -     */
	    ; {
		UR->DataType = NDDT_RAW;
		UR->Data.Raw.Len = DLength;
		UR->Data.Raw.Ofs = SavedPos - Packed;
	    }
	    break;
	}
	/*   */
	UR->NextRec = U->Records;
	U->Records = UR;
	UR = NULL;
	/* ... and to the next record */
	Pos = SavedPos + DLength;
    }

    /*    */

    ; {
	ND_UR *UR1, *UR2, *UR3;

	UR1 = U->Records;
	UR2 = NULL;
	while (UR1) {
	    UR3 = UR1->NextRec;
	    UR1->NextRec = UR2;
	    UR2 = UR1;
	    UR1 = UR3;
	}
	U->Records = UR2;
    }
    ier = 0;
    if (ierp)
	*ierp = ier;
    return U;

    /* Cleanup -   :     */

CU:
    if (ierp)
	*ierp = ier;
    if (SI) {
	if (SI->PrimaryNS) free(SI->PrimaryNS);
	if (SI->ZoneMaster) free(SI->ZoneMaster);
	free(SI);
    }
    if (MI) {
        if (MI->Server) free(MI->Server);
	free(MI);
    }
    if (UR) ND_UR_Free(UR);
    ND_Unpacked_Free(U);
    return NULL;
}
