/*-
 * Copyright (c) 2003 Valentin Nechayev <netch@netch.kiev.ua>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: verify.c,v 1.37 2003/09/16 13:56:09 netch Exp $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <time.h>
#include <unistd.h>
#include <assert.h>
#include <err.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifdef WITH_BIND_PORT
#include <bind/arpa/nameser.h>
#include <bind/netdb.h>
#include <bind/resolv.h>
#else
#include <arpa/nameser.h>
#include <netdb.h>
#include <resolv.h>
#endif
#include <assert.h>
#include <ctype.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#include "verify.h"
#include "nuxres.h"
#include "io3.h"
#include "mcom.h"
#include "avltree.h"

struct mx_item {
    TAILQ_ENTRY(mx_item) mxi_link;
    char *mxi_name;
    unsigned mxi_distance;
};

struct ipaddr_item {
    TAILQ_ENTRY(ipaddr_item) iat_link;
    unsigned long iat_addr;	/* IPv4 only */
    char *iat_hostname;	/* Name of host gave this address */
    unsigned iat_f1:1;
};

typedef TAILQ_HEAD(, ipaddr_item) ipaddr_list;

struct domain_entry {
    unsigned dce_magic;
    TAILQ_ENTRY(domain_entry) dce_link;
    char *dce_name;
    int dce_cclass;
    int dce_herror;	/* from netdb */
    char *dce_msg;
    time_t dce_time;
    time_t dce_expire;
    TAILQ_HEAD(, mx_item) dce_mxes;
    TAILQ_HEAD(, ipaddr_item) dce_ips;
};

#define DCE_MAGIC 0xdafa0077

pthread_mutex_t addrcache_mx;
AVL_IX_DESC addrcache_byaddr;
AVL_IX_DESC addrcache_byexpire;
TAILQ_HEAD(, address_entry) addrcache_list;
int addrcache_len = 0;

pthread_mutex_t domcache_mx;
AVL_IX_DESC domcache_bydom;
AVL_IX_DESC domcache_byexpire;
TAILQ_HEAD(, domain_entry) domcache_list;
int domcache_len = 0;

pthread_mutex_t extmap_mx;
char *extmap_path;
AVL_IX_DESC extmap;
time_t extmap_opened;
const int pass_code = 2;
const int tmpfail_code = 4;
const int reject_code = 5;

int bestmxonly = 0;
int notnullrevrej = 0;

#define RBUFSIZE 65536

/* res_state cache element */

typedef struct s_rescache_entry {
    TAILQ_ENTRY(s_rescache_entry) sr_link;
    struct __res_state sr_res;
    time_t sr_expire;
    unsigned char sr_buf[RBUFSIZE];
} rescache_entry;

pthread_mutex_t rescache_mx;
TAILQ_HEAD(, s_rescache_entry) rescache;

void verify_init(const char *);
int verify_address(char *, struct address_entry **);

static int connect_timed(int, const struct sockaddr *, socklen_t);
static int do_proto(const char *, struct address_entry **, io3 *, io3 *, const char *, const char *);
static int smtp_get_reply(io3 * fi, char *ibuf, size_t ibufsize);
static int rtrim_crlf(char *ibuf, size_t ngot);

static void addrcache_flush_head(void);
static void addrcache_xadd_copy(const struct address_entry *);
static int addrcache_expiration_key(char *, size_t, const struct address_entry *);

static int domain_lookup(const char *, struct domain_entry **, ipaddr_list *);
static int domain_lookup_add_ips(const char *, ipaddr_list *);

static void domcache_flush_head(void);
static void domcache_xadd_copy(const struct domain_entry *);
static int domcache_expiration_key(char *, size_t, const struct domain_entry *);

static int lookup_mx(const char *, struct domain_entry **);
static int lookup_mx_direct(const char *, struct domain_entry **);

static int lookup_ip(const char *, struct domain_entry **);
static int lookup_ip_direct(const char *, struct domain_entry **);

static char *xbracketizedup(const char *);

static rescache_entry *get_rescache_entry(void);
static void put_rescache_entry(rescache_entry *);

static struct address_entry *new_address_entry(void);
static void address_entry_mladd(struct address_entry *, const char *,...);
static struct address_entry *address_entry_xclone(const struct address_entry *);
void address_entry_free(struct address_entry *);

static struct domain_entry *new_domain_entry(void);
static void domain_entry_free(struct domain_entry *);
static struct domain_entry *domain_entry_xclone(const struct domain_entry *);

static void free_ipaddrs(ipaddr_list *);
static void lc(char **);	/* lowercase string */
static void extmap_read(void);

int check_extmap(char *, unsigned long);

/* Initialize subsystem. */

void
verify_init(const char *MapPath)
{
    int rc;

    if ((rc = pthread_mutex_init(&addrcache_mx, NULL)) != 0)
	errc(1, rc, "pthread_mutex_init()");
    TAILQ_INIT(&addrcache_list);
    avl_create_index(&addrcache_byaddr, AVL_NO_DUP_KEYS, 0);
    avl_create_index(&addrcache_byexpire, AVL_NO_DUP_KEYS, 0);
    addrcache_len = 0;
    if ((rc = pthread_mutex_init(&domcache_mx, NULL)) != 0)
	errc(1, rc, "pthread_mutex_init()");
    TAILQ_INIT(&domcache_list);
    avl_create_index(&domcache_bydom, AVL_NO_DUP_KEYS, 0);
    avl_create_index(&domcache_byexpire, AVL_NO_DUP_KEYS, 0);
    domcache_len = 0;
    if ((rc = pthread_mutex_init(&rescache_mx, NULL)) != 0)
	errc(1, rc, "pthread_mutex_init()");
    TAILQ_INIT(&rescache);
    extmap_path = MapPath ? xstrdup(MapPath) : NULL;
    if (MapPath != NULL) {
	avl_create_index(&extmap, AVL_NO_DUP_KEYS, 0);
	extmap_read();
	if ((rc = pthread_mutex_init(&extmap_mx, NULL)) != 0)
	    errc(1, rc, "pthread_mutex_init()");
	time(&extmap_opened);
    }
}

/*
 * Main entry. Checks email address <address>. Pass result in *a. Result must
 * be destroyed by caller using address_entry_free().
 */

int
verify_address(char *address, struct address_entry **a)
{
    int rc, f;
    char pebuf[sizeof(AVL_IX_REC)+2000];
    AVL_IX_REC *pe = (AVL_IX_REC *) pebuf;
    char *domain;
    struct address_entry *ace = NULL;
    struct domain_entry *dce = NULL;
    struct ipaddr_item *ipaddr;
    int cnt_conn;
    ipaddr_list ipaddrs;
    time_t now;

    assert(a);
    *a = NULL;
    time(&now);
    lc(&address);

    /* Lookup address in external map, if given. */

    if (extmap_path) {

	sprintf(pe->key, "from:%s", address);

	rc = pthread_mutex_lock(&extmap_mx);
	if (rc != 0)
	    errc(1, rc, "pthread_mutex_lock[%d]", __LINE__);

	if (difftime(now, extmap_opened) >= 60) {
	    extmap_read();
	    time(&extmap_opened);
	}
	if (avl_find_key(pe, &extmap) == AVL_IX_OK) {
	    pthread_mutex_unlock(&extmap_mx);
	    ace = new_address_entry();
	    ace->ace_addr = xstrdup(address);
	    ace->ace_cclass = *(int *)pe->recptr;
	    ace->ace_msg = xstrdup("By Administrator");
	    ace->ace_time = time(NULL);
	    ace->ace_expire = ace->ace_time;
	    ace->ace_magic = ACE_MAGIC;
	    *a = ace;
	    return 1;
	}
	pthread_mutex_unlock(&extmap_mx);
    }
    /* Try to get cached address data. */
    rc = pthread_mutex_lock(&addrcache_mx);
    if (rc != 0)
	errc(1, rc, "pthread_mutex_lock[%d]", __LINE__);
    addrcache_flush_head();
    strcpy(pe->key, address);
    if (avl_find_key(pe, &addrcache_byaddr) == AVL_IX_OK) {
	/*
	 * Found in cached data. Return clone. As addrcache_flush_head() was
	 * called already, we don't need here to check for expiration.
	 */
	ace = (struct address_entry *) pe->recptr;
	assert(ace);
	assert(ace->ace_magic == ACE_MAGIC);
	*a = address_entry_xclone(ace);
	pthread_mutex_unlock(&addrcache_mx);
	return 1;
    }
    pthread_mutex_unlock(&addrcache_mx);
    /*
     * Do domain lookup. If it succeeds, continue with address. Otherwise,
     * copy domain status.
     */
    domain = strrchr(address, '@');
    if (!domain) {
	return -1;		/* XXX */
    }
    domain++;	/* now domain points to domain part of address */
    if (extmap_path) {
	size_t dl;

	sprintf(pe->key, "fromdom:%s", domain);

	dl = strlen(pe->key);
	if (*(pe->key) && pe->key[dl-1] == '.') {
	    dl--;
	    pe->key[dl] = 0;
	}
	rc = pthread_mutex_lock(&extmap_mx);
	if (rc != 0)
	    errc(1, rc, "pthread_mutex_lock[%d]", __LINE__);
	if (avl_find_key(pe, &extmap) == AVL_IX_OK) {
	    pthread_mutex_unlock(&extmap_mx);
	    ace = new_address_entry();
	    ace->ace_addr = xstrdup(address);
	    ace->ace_cclass = *(int *)pe->recptr;
	    ace->ace_msg = xstrdup("By Administrator");
	    ace->ace_time = time(NULL);
	    ace->ace_expire = ace->ace_time;
	    ace->ace_magic = ACE_MAGIC;
	    *a = ace;
	    return 1;
	}
	pthread_mutex_unlock(&extmap_mx);
    }
    if (strlen(domain) > 255) {
	return -1;		/* XXX */
    }
    TAILQ_INIT(&ipaddrs);
    rc = domain_lookup(domain, &dce, &ipaddrs);
    /*
     * domain_lookup()   .      -
     *    .    [best]MX'.
     */
    f = 0;
    if (rc < 0 || !dce)
	f = 1;
    if (!f && dce && dce->dce_herror == HOST_NOT_FOUND) {
	domain_entry_free(dce);
	dce = NULL;
	ace = new_address_entry();
	ace->ace_addr = xstrdup(address);
	ace->ace_cclass = 5;
	address_entry_mladd(ace, "Sender verify callout failed: "
				"Nonexistent domain");
	address_entry_mladd(ace, "Sender address: <%s>", address);
	ace->ace_time = time(NULL);
	ace->ace_expire = ace->ace_time + 3600;	/* XXX      */
	ace->ace_magic = ACE_MAGIC;
	addrcache_xadd_copy(ace);
	*a = ace;
	free_ipaddrs(&ipaddrs);
	return 1;
    }
    if (dce && (dce->dce_cclass == 5 || dce->dce_cclass == 4)) {
	int cc = dce->dce_cclass;

	domain_entry_free(dce);
	dce = NULL;
	ace = new_address_entry();
	ace->ace_addr = xstrdup(address);
	ace->ace_cclass = cc;
	ace->ace_time = time(NULL);
	address_entry_mladd(ace, "Sender verify callout failed: "
				"Unrecoverable sender domain lookup error");
	address_entry_mladd(ace, "Sender address: <%s>", address);
	ace->ace_expire = ace->ace_time + 3600;	/* XXX      */
	ace->ace_magic = ACE_MAGIC;
	addrcache_xadd_copy(ace);
	*a = ace;
	free_ipaddrs(&ipaddrs);
	return 1;
    }
    if (dce &&
	(dce->dce_herror == TRY_AGAIN || dce->dce_herror == NO_RECOVERY))
	    f = 1;
    if (TAILQ_EMPTY(&ipaddrs)) f = 1;
    if (f) {
	/*
	 *    "      ".
	 *   ,    .
	 */
	domain_entry_free(dce);
	dce = NULL;
	ace = new_address_entry();
	ace->ace_addr = xstrdup(address);
	ace->ace_cclass = 4;
	address_entry_mladd(ace, "Sender verify callout failed: "
				"Transient domain lookup error");
	address_entry_mladd(ace, "Sender address: <%s>", address);
	ace->ace_time = time(NULL);
	ace->ace_expire = ace->ace_time + 600;	/* XXX 10 min */
	ace->ace_magic = ACE_MAGIC;
	addrcache_xadd_copy(ace);
	*a = ace;
	free_ipaddrs(&ipaddrs);
	return 1;
    }
    /* - ... */
    assert(dce);
    /*  bestMXes... */
    cnt_conn = 0;
    TAILQ_FOREACH(ipaddr, &ipaddrs, iat_link) {
	ipaddr->iat_f1 = 0;
    }
    TAILQ_FOREACH(ipaddr, &ipaddrs, iat_link) {
	int si, so, rc_proto;
	struct ipaddr_item *ip2;
	io3 *fi = NULL;
	io3 *fo = NULL;
	struct sockaddr_in sia;
	char ipbuf[16];

	/* Eliminate duplicates... */
	if (ipaddr->iat_f1)
	    continue;
	for (ip2 = TAILQ_NEXT(ipaddr, iat_link); ip2 != NULL;
	     ip2 = TAILQ_NEXT(ip2, iat_link)) {
	    assert(ip2 != ipaddr);
	    if (ip2->iat_addr == ipaddr->iat_addr)
		ip2->iat_f1 = 1;
	}
	/* gul: check for fake MX-es (IN A 0.0.0.0 or 127.x.x.x) */
	/* db: added 10/8, 172.16/12, 192.168/16, 169.254/16, 224/8, 232/8 */
	if (ipaddr->iat_addr == 0 || 
			(ipaddr->iat_addr & 0xff000000) == 0x7f000000 ||
			(ipaddr->iat_addr & 0xff000000) == 0x0a000000 ||
			(ipaddr->iat_addr & 0xff000000) == 0xe0000000 ||
			(ipaddr->iat_addr & 0xff000000) == 0xe8000000 ||
			(ipaddr->iat_addr & 0xfff00000) == 0xac100000 ||
			(ipaddr->iat_addr & 0xffff0000) == 0xc0a80000 ||
			(ipaddr->iat_addr & 0xffff0000) == 0xa9fe0000) {
	    struct in_addr mxaddr;

	    mxaddr.s_addr = htonl(ipaddr->iat_addr);
	    inet_ntop(AF_INET, &mxaddr, ipbuf, sizeof ipbuf);
	    ace = new_address_entry();
	    ace->ace_addr = xstrdup(address);
	    ace->ace_cclass = 5;
	    address_entry_mladd(ace,
			    "Sender verify callout failed: bad MX ip %s",
				ipbuf);
	    address_entry_mladd(ace, "Sender address: <%s>", address);
	    ace->ace_time = time(NULL);
	    ace->ace_expire = ace->ace_time + 3600;	/* XXX 1 hour */
	    ace->ace_magic = ACE_MAGIC;
	    break;
	}
	/*
	 * Try to connect. If can't establish dialog with any (including
	 * connection problems and protocol initiation problems), mark the
	 * whole domain having transient error. If have result on any bestMX,
	 * record this result.
	 */
	si = socket(PF_INET, SOCK_STREAM, 0);
	if (si < 0)
	    err(1, "socket()");
	bzero(&sia, sizeof sia);
	sia.sin_family = AF_INET;
	sia.sin_addr.s_addr = htonl(ipaddr->iat_addr);
	sia.sin_port = htons(25);
	if (connect_timed(si, (struct sockaddr *) & sia, sizeof sia) < 0) {
	    close(si);
	    continue;
	}
	/* Connected. Create streams and talk */
	rc = fcntl(si, F_GETFL, 0);
	if (rc < 0)
	    err(1, "F_GETFL");
	rc = fcntl(si, F_SETFL, rc | O_NONBLOCK);
	if (rc < 0)
	    err(1, "F_SETFL");
	so = dup(si);
	if (so < 0)
	    err(1, "dup()");
	++cnt_conn;
	fi = io3_fdopen(si, "r");
	if (!fi)
	    err(1, "io3_fdopen()");
	si = -1;
	fo = io3_fdopen(so, "a");
	if (!fo)
	    err(1, "io3_fdopen()");
	so = -1;
	strlcpy(ipbuf, inet_ntoa(sia.sin_addr), sizeof ipbuf);
	rc_proto = do_proto(address, &ace, fi, fo, ipaddr->iat_hostname, ipbuf);
	if (si != -1)
	    close(si);
	if (so != -1)
	    close(so);
	if (fi)
	    io3_close(fi);
	if (fo)
	    io3_close(fo);
	if (rc_proto != 0)
	    break;
    }
    free_ipaddrs(&ipaddrs);
    if (!ace) {

	/*    .  temporary failure. */
	domain_entry_free(dce);
	dce = NULL;
	ace = new_address_entry();
	ace->ace_addr = xstrdup(address);
	ace->ace_cclass = 4;
	if (bestmxonly)
		address_entry_mladd(ace, "Sender verify callout failed: "
				"can't connect to primary MX");
	else
		address_entry_mladd(ace, "Sender verify callout failed: "
				"can't connect to any MX");
	address_entry_mladd(ace, "Sender address: <%s>", address);
	ace->ace_time = time(NULL);
	ace->ace_expire = ace->ace_time + 600;	/* XXX 10 min */
	ace->ace_magic = ACE_MAGIC;
	addrcache_xadd_copy(ace);
	*a = ace;
	return 1;
    }
    assert(ace);
    addrcache_xadd_copy(ace);
    *a = ace;
    return 1;
}

static int
connect_timed(int fd, const struct sockaddr *a, socklen_t alen)
{
    int r0, r, se;
    struct timeval tv0, tv1, tv2;

    r = fcntl(fd, F_GETFL, 0);
    if (r < 0)
	return -1;
    r0 = r;
    r = fcntl(fd, F_SETFL, r0 | O_NONBLOCK);
    if (r < 0)
	return -1;
    r = connect(fd, a, alen);
    if (r == 0 || (r < 0 && errno != EINPROGRESS)) {
	se = errno;
	fcntl(fd, F_SETFL, r0);
	errno = se;
	return r;
    }
    fcntl(fd, F_SETFL, r0);
    gettimeofday(&tv0, NULL);
    tv1 = tv0;
    tv1.tv_sec += 120;		/* 2 min timeout */
    assert(fd < FD_SETSIZE);
    for (;;) {
	fd_set fds;

	gettimeofday(&tv0, NULL);
	io3_tv_sub3(&tv1, &tv0, &tv2);
	FD_ZERO(&fds);
	FD_SET(fd, &fds);
	r = select(fd + 1, NULL, &fds, NULL, &tv2);
	if (r > 0)
	    break;
	gettimeofday(&tv0, NULL);
	if (io3_tv_compare(&tv1, &tv0) <= 0) {
	    errno = ETIMEDOUT;
	    return -1;
	}
    }
    return 0;
}

static int
do_proto(const char *address, struct address_entry **a, io3 *fi, io3 *fo, const char *domain, const char *ip)
{
    int rc;
    struct timeval tv;
    char ibuf[2049], ibuf2[2049], obuf[2049], obufq[2049], hostnamebuf[256];
    char *p;
    size_t ll;
    ssize_t nput;
    struct address_entry *ace;

    gettimeofday(&tv, NULL);
    tv.tv_sec += 300;		/* 5      */
    io3_set_timeout(fi, &tv);
    io3_set_timeout(fo, &tv);
    /* Get banner */
    rc = smtp_get_reply(fi, ibuf, sizeof ibuf);
    if (!rc) return 0;
    if (ibuf[0] == '5')	{
	    obuf[0] = 0;
	    goto R;
    }
    if (ibuf[0] == '4')	return 0;
    if (ibuf[0] != '2')	return 0;
    /* Say HELO */
    if (gethostname(hostnamebuf, sizeof hostnamebuf) < 0)
	err(1, "gethostname()");
    ll = snprintf(obuf, sizeof obuf, "HELO %s\r\n", hostnamebuf);
    if (ll >= sizeof obuf) return 0;
    nput = io3_write(fo, obuf, strlen(obuf));
    if (io3_timed_out(fo) || nput != ll) return 0;
    rc = smtp_get_reply(fi, ibuf, sizeof ibuf);
    if (!rc) return 0;
    /* MAIL FROM: */
    ll = snprintf(obuf, sizeof obuf, "MAIL FROM:<>\r\n");
    if (ll >= sizeof obuf) return 0;
    nput = io3_write(fo, obuf, strlen(obuf));
    if (io3_timed_out(fo) || nput != ll) return 0;
    rc = smtp_get_reply(fi, ibuf, sizeof ibuf);
    if (!rc) return 0;
    /*
     * This is most discussible place. Some systems simply reject empty
     * reverse address. This is ugly violation of RFC821, but this is
     * widespread behavior, so we shall have configurable switch to detect
     * such cases. Really, seen 5xx as reply to 'mail from:', we have only
     * two variants: reject, or accept. Tempfail isn't good case.
     */
    if (ibuf[0] == '5') {
	    if (notnullrevrej)
		    goto R;
	    goto Accept;
    }
    if (ibuf[0] != '2')	return 0;
    /* RCPT TO: */
    ll = snprintf(obuf, sizeof obuf, "RCPT TO:<%s>\r\n", address);
    if (ll >= sizeof obuf) return 0;
    nput = io3_write(fo, obuf, strlen(obuf));
    if (io3_timed_out(fo) || nput != ll) return 0;
    rc = smtp_get_reply(fi, ibuf, sizeof ibuf);
    if (!rc) return 0;
    /* QUIT */
    ll = snprintf(obufq, sizeof obufq, "QUIT\r\n");
    if (ll >= sizeof obufq) goto S;
    nput = io3_write(fo, obufq, strlen(obufq));
    if (!io3_timed_out(fo) && nput == ll)
	(void)smtp_get_reply(fi, ibuf2, sizeof ibuf2);
    /* Save results */
S:
    if (ibuf[0] == '2') {
Accept:
	ace = new_address_entry();
	ace->ace_addr = xstrdup(address);
	ace->ace_cclass = 2;
	ace->ace_msg = NULL;
	ace->ace_time = time(NULL);
	ace->ace_expire = ace->ace_time + 3600;	/* XXX 1 hour */
	ace->ace_magic = ACE_MAGIC;
	*a = ace;
	return 1;
    }
R:
    ace = new_address_entry();
    ace->ace_addr = xstrdup(address);
    ace->ace_cclass = ibuf[0] - '0';
    address_entry_mladd(ace, "Sender verify callout failed: no route back");
    address_entry_mladd(ace, "Sender address: <%s>", address);
    address_entry_mladd(ace, "MX used: %s [%s]", domain, ip);
    if ((p = strchr(obuf, '\r'))) *p = 0;
    if ((p = strchr(obuf, '\n'))) *p = 0;
    if (obuf[0]) {
	    address_entry_mladd(ace, "Client command: %s", obuf);
	    address_entry_mladd(ace, "Server response: %s", ibuf);
    } else
	    address_entry_mladd(ace, "Server banner: %s", ibuf);
    ace->ace_time = time(NULL);
    ace->ace_expire = ace->ace_time + 3600;	/* XXX 1 hour */
    ace->ace_magic = ACE_MAGIC;
    *a = ace;
    return 1;
}

static int
smtp_get_reply(io3 *fi, char *ibuf, size_t ibufsize)
{
    int seen_lf;

    for (;;) {
	ssize_t ngot = io3_fngets(fi, ibuf, ibufsize, &seen_lf);

	if (io3_timed_out(fi) || ngot <= 0 || !seen_lf) return 0;
	if (!rtrim_crlf(ibuf, (size_t) ngot)) return 0;
	/*
	 *         
	 *  
	 */
	if (!isdigit(ibuf[0]))
	    continue;
	if (strlen(ibuf) < 3)
	    continue;
	if (ibuf[0] < '2' || ibuf[0] > '5')
	    continue;
	if (!isdigit(ibuf[1]) || !isdigit(ibuf[2]))
	    continue;
	if (ibuf[3] == '-')
	    continue;
	if (ibuf[0] == '2' || ibuf[0] == '4' || ibuf[0] == '5')
	    break;
    }
    return 1;			/* got it */
}

static int
rtrim_crlf(char *ibuf, size_t ngot)
{
    size_t n1;
    char *z;

    if (ngot == 0)
	return 0;
    while ((z = memchr(ibuf, 0, ngot)))	*z = ' ';
    n1 = ngot - 1;
    /* We allow both LF or CRLF */
    if (ibuf[n1] != '\n')
	return 0;
    ibuf[n1] = 0;
    if (n1 == 0)
	return 1;
    --n1;
    if (ibuf[n1] == '\r')
	ibuf[n1] = 0;
    return 1;
}

/* Flush cache. Call only with lock owned by us! */

static void
addrcache_flush_head(void)
{
    time_t now;
    struct address_entry *p;
    char pebuf[AVL_REC_LEN];
    AVL_IX_REC *pe = (AVL_IX_REC *) pebuf;

    /*
     *      .  
     *   .
     */

    time(&now);
    while ((p = TAILQ_FIRST(&addrcache_list)) && p->ace_expire < now) {
	char buf[2000];

	TAILQ_REMOVE(&addrcache_list, p, ace_link);
	strcpy(pe->key, p->ace_addr);
	if (avl_delete_key(pe, &addrcache_byaddr) != AVL_IX_OK)
	    err(1, "avltree::del()");
	addrcache_expiration_key(buf, sizeof buf, p);
	strcpy(pe->key, buf);
	if (avl_delete_key(pe, &addrcache_byexpire) != AVL_IX_OK)
	    err(1, "avltree::del()");
	address_entry_free(p);
	--addrcache_len;
    }
}

static void
addrcache_xadd_copy(const struct address_entry *s)
{
    struct address_entry *n;
    struct address_entry *a2;
    char pebuf[AVL_REC_LEN];
    AVL_IX_REC *pe = (AVL_IX_REC *) pebuf;
    int rc;
    char buf[2000];

    n = address_entry_xclone(s);
    /*
     * Put by address. If record for this address already exists, another
     * thread raced us. This is almost normal. In this case, drop our copy as
     * too late.
     */
    strcpy(pe->key, n->ace_addr);
    pe->recptr = (void *)n;
    rc = pthread_mutex_lock(&addrcache_mx);
    if (rc != 0)
	errc(1, rc, "pthread_mutex_lock[%d]", __LINE__);
    if (avl_add_key(pe, &addrcache_byaddr) != AVL_IX_OK) {
	/* We were raced, skip it */
	pthread_mutex_unlock(&addrcache_mx);
	address_entry_free(n);
	return;
    }
    addrcache_expiration_key(buf, sizeof buf, n);
    /* Add to cached data list, sorted by expiration time. */
    if (TAILQ_EMPTY(&addrcache_list))
	TAILQ_INSERT_TAIL(&addrcache_list, n, ace_link);
    else {
	strcpy(pe->key, buf);
	/* Find position in buffer... */
	if (avl_locate_key(pe, &addrcache_byexpire) != AVL_EOIX) {
	    /*
	     * There is smallest key which is equal or greater. Insert before
	     * it.
	     */
	    a2 = (struct address_entry *) pe->recptr;
	    assert(a2->ace_magic == ACE_MAGIC);
	    TAILQ_INSERT_BEFORE(a2, n, ace_link);
	} else {
	    /* No key equal or greater, insert to tail */
	    TAILQ_INSERT_TAIL(&addrcache_list, n, ace_link);
	}
    }
    /* Add to expiration index. It must be added successfully. */
    strcpy(pe->key, buf);
    pe->recptr = (void *)n;
    if (avl_add_key(pe, &addrcache_byexpire) != AVL_IX_OK)
	errx(1, "addrcache_add_copy(): error putting to addrcache_byexpire");
    ++addrcache_len;
    pthread_mutex_unlock(&addrcache_mx);
}

static int
addrcache_expiration_key(char *buf, size_t buflen, const struct address_entry *ace)
{
    return snprintf(buf, buflen, "%010lX:%s",
		    (unsigned long)ace->ace_expire, ace->ace_addr);
}

/*
 *    ,  ,  ,    DNS.
 *     ,   MX'.
 */

static int
domain_lookup(const char *domain0, struct domain_entry **p, ipaddr_list *ipaddrs)
{
    int rc;
    char *dom_nolit = NULL;
    char *domain;
    int flag_nomx = 0;

    assert(domain0);
    assert(*domain0);
    *p = NULL;

    if (*domain0 == '[') {
	const char *p1 = strrchr(domain0 + 1, ']');
	size_t l1;

	if (p1)
	    l1 = p1 - (domain0 + 1);
	else
	    l1 = strlen(domain0 + 1);
	domain = dom_nolit = xstrndup(domain0 + 1, l1);
	flag_nomx = 1;
    } else
	domain = (char *)domain0;

    /* Try to get MXes. */
    if (!flag_nomx) {
	rc = lookup_mx(domain, p);
	if (rc < 0) {
	    free(dom_nolit);
	    return rc;
	} else if (rc == 0)
	    flag_nomx = 1;
	else {
	    struct mx_item *mx;
	    struct domain_entry *d = *p;

	    assert(d);
	    assert(d->dce_magic == DCE_MAGIC);
	    if (!TAILQ_FIRST(&d->dce_mxes)) {
		/* Cached case: no data */
		flag_nomx = 1;
		goto nomx;
	    }
	    /* We see MXes. Iterate them. */
	    TAILQ_FOREACH(mx, &d->dce_mxes, mxi_link) {
		domain_lookup_add_ips(mx->mxi_name, ipaddrs);
	    }
	    free(dom_nolit);
	    return 1;
	}
    }
nomx:
    /* We should provide some simple reply */
    rc = lookup_ip(domain, p);
    if (rc > 0) {
	/* Copy IPs to ipaddrs */
	struct ipaddr_item *i;

	assert(p);
	TAILQ_FOREACH(i, &(*p)->dce_ips, iat_link) {
	    struct ipaddr_item *in;
	    in = xcalloc(1, sizeof(*in));
	    in->iat_addr = i->iat_addr;
	    in->iat_hostname = xstrdup(i->iat_hostname);
	    TAILQ_INSERT_TAIL(ipaddrs, in, iat_link);
	}
    }
    free(dom_nolit);
    return rc;
}

static int
domain_lookup_add_ips(const char *domain, ipaddr_list *ipaddrs)
{
    int rc;
    struct domain_entry *p = NULL;
    struct ipaddr_item *i;

    rc = lookup_ip(domain, &p);
    if (rc <= 0)
	return rc;
    TAILQ_FOREACH(i, &p->dce_ips, iat_link) {
	struct ipaddr_item *in;

	in = xcalloc(1, sizeof(*in));
	in->iat_addr = i->iat_addr;
	in->iat_hostname = xstrdup(i->iat_hostname);
	TAILQ_INSERT_TAIL(ipaddrs, in, iat_link);
    }
    domain_entry_free(p);
    return 1;
}

/* Flush cache. Call only with lock owned by us! */

static void
domcache_flush_head(void)
{
    time_t now;
    struct domain_entry *p;
    char pebuf[AVL_REC_LEN];
    AVL_IX_REC *pe = (AVL_IX_REC *) pebuf;

    /*
     *      .
     *     .
     */

    time(&now);
    while ((p = TAILQ_FIRST(&domcache_list)) != NULL &&
	   p->dce_expire < now) {
	char buf[2000];

	TAILQ_REMOVE(&domcache_list, p, dce_link);
	strcpy(pe->key, p->dce_name);
	if (avl_delete_key(pe, &domcache_bydom) != AVL_IX_OK)
	    err(1, "avltree::del()");
	domcache_expiration_key(buf, sizeof buf, p);
	strcpy(pe->key, buf);
	if (avl_delete_key(pe, &domcache_byexpire) != AVL_IX_OK)
	    err(1, "avltree::del()");
	domain_entry_free(p);
	--domcache_len;
    }
}

static void
domcache_xadd_copy(const struct domain_entry *s)
{
    struct domain_entry *n;
    struct domain_entry *d2;
    char pebuf[AVL_REC_LEN];
    AVL_IX_REC *pe = (AVL_IX_REC *) pebuf;
    int rc;
    char buf[2000];

    n = domain_entry_xclone(s);
    strcpy(pe->key, n->dce_name);
    pe->recptr = (void *)n;
    rc = pthread_mutex_lock(&domcache_mx);
    if (rc != 0)
	errc(1, rc, "pthread_mutex_lock[%d]", __LINE__);
    if (avl_add_key(pe, &domcache_bydom) != AVL_IX_OK) {
	pthread_mutex_unlock(&domcache_mx);
	domain_entry_free(n);
	return;
    }
    domcache_expiration_key(buf, sizeof buf, n);
    if (TAILQ_EMPTY(&domcache_list))
	TAILQ_INSERT_TAIL(&domcache_list, n, dce_link);
    else {
	strcpy(pe->key, buf);
	/* Find position in buffer... */
	if (avl_locate_key(pe, &domcache_byexpire) != AVL_EOIX) {
	    /*
	     * There is smallest key which is equal or greater. Insert before
	     * it.
	     */
	    d2 = (struct domain_entry *) pe->recptr;
	    assert(d2->dce_magic == DCE_MAGIC);
	    TAILQ_INSERT_BEFORE(d2, n, dce_link);
	} else {
	    /* No key equal or greater, insert to tail */
	    TAILQ_INSERT_TAIL(&domcache_list, n, dce_link);
	}
    }
    strcpy(pe->key, buf);
    pe->recptr = (void *)n;
    if (avl_add_key(pe, &domcache_byexpire) != AVL_IX_OK)
	errx(1, "domcache_xadd_copy(): error putting to domcache_byexpire");
    ++domcache_len;
    pthread_mutex_unlock(&domcache_mx);

}

static int
domcache_expiration_key(char *buf, size_t buflen, const struct domain_entry *d)
{
    return snprintf(buf, buflen, "%010lX:%s",
		    (unsigned long)d->dce_expire, d->dce_name);
}

/* Find only MXes. Deal with cache. */

static int
lookup_mx(const char *domain, struct domain_entry **p)
{
    int rc;
    char pebuf[AVL_REC_LEN];
    AVL_IX_REC *pe = (AVL_IX_REC *) pebuf;

    assert(p);
    if (*p) domain_entry_free(*p);
    *p = NULL;

    assert(domain);
    if (!domain[0] || domain[0] == '[')
	return 0;
    /* Try to find in cache... */
    rc = pthread_mutex_lock(&domcache_mx);
    if (rc != 0)
	errc(1, rc, "pthread_mutex_lock[%d]", __LINE__);
    domcache_flush_head();
    strcpy(pe->key, domain);
    if (avl_find_key(pe, &domcache_bydom) == AVL_IX_OK) {
	/* -  */
	struct domain_entry *dce;

	dce = (struct domain_entry *) pe->recptr;
	assert(dce);
	assert(dce->dce_magic == DCE_MAGIC);
	*p = domain_entry_xclone(dce);
	pthread_mutex_unlock(&domcache_mx);
	return 1;
    }
    pthread_mutex_unlock(&domcache_mx);
    rc = lookup_mx_direct(domain, p);
    if (rc > 0) {
	assert(*p);
	domcache_xadd_copy(*p);
    }
    return rc;
}

static int
lookup_mx_direct(const char *domain, struct domain_entry **p)
{
    int l1, nseen, minprec, rc, rs_err;
    rescache_entry *re;
    struct domain_entry *dce = NULL;
    ND_UR *UR = NULL;
    ND_Unpacked *U1 = NULL;
    HEADER Hr;
    unsigned ttl = UINT_MAX;

    assert(p);
    if (*p) domain_entry_free(*p);
    *p = NULL;
    re = get_rescache_entry();
    assert(re);
    /*
     * res_nquery() returns -1 and sets h_errno if it sees an error in
     * reply->rcode, or if reply->ancount==0 (in this case it sets
     * h_errno:=NO_DATA). So, we should rely here on its codes. h_errno is
     * set both global and in res_state.
     */
    l1 = res_nquery(&re->sr_res, domain, C_IN, T_MX, re->sr_buf, RBUFSIZE);
    if (re->sr_res.options & RES_DEBUG)
	fflush(NULL);
    rs_err = re->sr_res.res_h_errno;
    if (l1 >= HFIXEDSZ) {
	memmove(&Hr, re->sr_buf, sizeof Hr);
	U1 = ND_Unpack(&rc, re->sr_buf, (size_t) l1);
    }
    put_rescache_entry(re);
    re = NULL;
    if ((l1 >= 0 && l1 < NS_HFIXEDSZ) ||
	(l1 == -1 && rs_err == SERVFAIL))
	goto ServFail;
    if (l1 >= HFIXEDSZ) {
	if (!U1)
	    return 0;
    }
    /* No domain is one simple case. */
    if ((l1 == -1 && rs_err == HOST_NOT_FOUND) ||
	(U1 != NULL && U1->rcode == NXDOMAIN)) {
	if (U1)
	    ND_Unpacked_Free(U1);
	dce = new_domain_entry();
	dce->dce_name = xstrdup(domain);
	dce->dce_cclass = 5;
	dce->dce_herror = HOST_NOT_FOUND;
	dce->dce_msg = xstrdup("Host not found");
	dce->dce_time = time(NULL);
	dce->dce_expire = dce->dce_time + 3600;	/* XXX fixed 1h */
	dce->dce_magic = DCE_MAGIC;
	*p = dce;
	return 1;
    }
    /*   NO_DATA   ... */
    if (l1 == -1 && rs_err == NO_DATA) {
	ND_Unpacked_Free(U1);
	return 0;
    }
    /*
     *       .   
     *       .
     */
    if (l1 == -1 && rs_err != NO_DATA) {
ServFail:
	if (U1)
	    ND_Unpacked_Free(U1);
	dce = new_domain_entry();
	dce->dce_name = xstrdup(domain);
	dce->dce_cclass = 4;
	dce->dce_herror = TRY_AGAIN;
	dce->dce_time = time(NULL);
	dce->dce_expire = dce->dce_time + 600;	/* XXX fixed 10m */
	dce->dce_magic = DCE_MAGIC;
	*p = dce;
	return 1;
    }
    /*  NOERROR.  . */
    assert(U1);
    nseen = 0;
    minprec = 65536;		/* More than maximum in 2 bytes */
    for (UR = U1->Records; UR != NULL; UR = UR->NextRec) {
	if (UR->Section != 2 || UR->Type != T_MX)
	    continue;
	++nseen;
	if (UR->Data.MX->Prec < minprec)
	    minprec = UR->Data.MX->Prec;
    }
    /*
     * nseen == 0   ,     MX .
     *       .  - ,   
     *    NO_DATA.
     *    -  0 ("  ")
     */
    if (nseen == 0) {
	ND_Unpacked_Free(U1);
	return 0;
    }
    dce = new_domain_entry();
    *p = dce;
    dce->dce_name = xstrdup(domain);
    dce->dce_cclass = 2;
    dce->dce_magic = DCE_MAGIC;
    dce->dce_time = time(NULL);
    for (UR = U1->Records; UR != NULL; UR = UR->NextRec) {
	if (UR->Section != 2 || UR->Type != T_MX)
	    continue;
	if (UR->Data.MX->Prec == minprec) {
	    struct mx_item *mx;

	    mx = xcalloc(1, sizeof(*mx));
	    mx->mxi_name = xstrdup(UR->Data.MX->Server);
	    mx->mxi_distance = UR->Data.MX->Prec;
	    TAILQ_INSERT_TAIL(&dce->dce_mxes, mx, mxi_link);
	    if (UR->Ttl < ttl)
		ttl = UR->Ttl;
	}
    }

    /* process bestmxonly flag */
    if (!bestmxonly) {
	    for (UR = U1->Records; UR != NULL; UR = UR->NextRec) {
		    if (UR->Section != 2 || UR->Type != T_MX)
			    continue;
		    if (UR->Data.MX->Prec > minprec && \
				    UR->Data.MX->Prec < 1000) {
			    struct mx_item *mx;

			    mx = xcalloc(1, sizeof(*mx));
			    mx->mxi_name = xstrdup(UR->Data.MX->Server);
			    mx->mxi_distance = UR->Data.MX->Prec;
			    TAILQ_INSERT_TAIL(&dce->dce_mxes, mx, mxi_link);
			    if (UR->Ttl < ttl)
				    ttl = UR->Ttl;
		    }
	    }
    }

    if (ttl > 86400) ttl = 86400;
    dce->dce_expire = dce->dce_time + ttl;
    ND_Unpacked_Free(U1);
    return 1;
}

static int
lookup_ip(const char *domain0, struct domain_entry **p)
{
    int rc;
    char *domain = NULL;
    char *d1 = NULL;
    char pebuf[AVL_REC_LEN];
    AVL_IX_REC *pe = (AVL_IX_REC *) pebuf;
    struct domain_entry* dce;
    struct in_addr iad;
    struct ipaddr_item* ia;

    assert(p);
    if (*p) domain_entry_free(*p);
    *p = NULL;

    assert(domain0);
    if (!*domain0)
	return 0;

    if (*domain0 == '[' ) {
	    char *d2;

	    d1 = xstrdup(domain0 + 1);
	    if ((d2 = strchr(d1, ']')) != 0) *d2 = 0;
	    rc = inet_aton(d1, &iad);
	    dce = new_domain_entry();
	    dce->dce_name = xstrdup(d1);
	    dce->dce_cclass = rc ? 2 : 5;
	    dce->dce_herror = 0;
	    dce->dce_msg = xstrdup("Host not found");
	    dce->dce_time = time(NULL);
	    dce->dce_expire = dce->dce_time;
	    dce->dce_magic = DCE_MAGIC;
	    *p = dce;
	    ia = xcalloc(1, sizeof(*ia));
	    ia->iat_addr = ntohl(iad.s_addr);
	    ia->iat_hostname = xstrdup(d1);
	    TAILQ_INSERT_TAIL(&dce->dce_ips, ia, iat_link);
	    if(!rc)
		    dce->dce_msg = xstrdup("550 Invalid domain literal");
	    free(d1);
	    return 1;
    }
    if (inet_aton(domain0, &iad)) {
	    dce = new_domain_entry();
	    dce->dce_name = xstrdup(domain0);
	    dce->dce_cclass = 2;
	    dce->dce_herror = 0;
	    dce->dce_time = time(NULL);
	    dce->dce_expire = dce->dce_time;
	    dce->dce_magic = DCE_MAGIC;
	    *p = dce;
	    ia = xcalloc(1, sizeof(*ia));
	    ia->iat_addr = ntohl(iad.s_addr);
	    ia->iat_hostname = xstrdup(domain0);
	    TAILQ_INSERT_TAIL(&dce->dce_ips, ia, iat_link);
	    return 1;
    }

    domain = d1 = xbracketizedup(domain0);

    /* Try to find in cache... */
    rc = pthread_mutex_lock(&domcache_mx);
    if (rc != 0)
	errc(1, rc, "pthread_mutex_lock[%d]", __LINE__);
    domcache_flush_head();
    strcpy(pe->key, domain);
    if (avl_find_key(pe, &domcache_bydom) == AVL_IX_OK) {
	/* -  */
	dce = (struct domain_entry *) pe->recptr;
	assert(dce);
	assert(dce->dce_magic == DCE_MAGIC);
	*p = domain_entry_xclone(dce);
	pthread_mutex_unlock(&domcache_mx);
	free(d1);
	return 1;
    }
    pthread_mutex_unlock(&domcache_mx);
    assert(*domain0 != '[');	/* XXX */
    rc = lookup_ip_direct(domain0, p);
    if (rc > 0) {
	assert(*p);
	domcache_xadd_copy(*p);
    }
    free(d1);
    return rc;
}

static int
lookup_ip_direct(const char *domain, struct domain_entry **p)
{
    int l1, nseen, rc, rs_err;
    rescache_entry *re;
    struct domain_entry *dce = NULL;
    ND_UR *UR = NULL;
    ND_Unpacked *U1 = NULL;
    HEADER Hr;
    unsigned ttl = UINT_MAX;

    assert(p);
    if (*p) domain_entry_free(*p);
    *p = NULL;
    re = get_rescache_entry();
    assert(re);
    /*
     * res_nquery() returns -1 and sets h_errno if it sees an error in
     * reply->rcode, or if reply->ancount==0 (in this case it sets
     * h_errno:=NO_DATA). So, we should rely here on its codes. h_errno is
     * set both global and in res_state.
     */
    l1 = res_nquery(&re->sr_res, domain, C_IN, T_A, re->sr_buf, RBUFSIZE);
    if (re->sr_res.options & RES_DEBUG)
	fflush(NULL);
    rs_err = re->sr_res.res_h_errno;
    if (l1 >= HFIXEDSZ) {
	memmove(&Hr, re->sr_buf, sizeof Hr);
	U1 = ND_Unpack(&rc, re->sr_buf, (size_t) l1);
    }
    put_rescache_entry(re);
    re = NULL;
    if ((l1 >= 0 && l1 < NS_HFIXEDSZ) ||
	(l1 == -1 && rs_err == SERVFAIL))
	goto ServFail;
    if (l1 >= HFIXEDSZ) {
	if (!U1)
	    return 0;
    }
    /* No domain is one simple case. */
    if ((l1 == -1 && rs_err == HOST_NOT_FOUND) ||
	(U1 != NULL && U1->rcode == NXDOMAIN)) {
ReturnNxdomain:
	if (U1)
	    ND_Unpacked_Free(U1);
	dce = new_domain_entry();
	dce->dce_name = xbracketizedup(domain);
	dce->dce_cclass = 5;
	dce->dce_herror = HOST_NOT_FOUND;
	dce->dce_msg = xstrdup("Host not found");
	dce->dce_time = time(NULL);
	dce->dce_expire = dce->dce_time + 3600;	/* XXX fixed 1h */
	dce->dce_magic = DCE_MAGIC;
	*p = dce;
	return 1;
    }
    /*   NO_DATA   ... */
    if (l1 == -1 && rs_err == NO_DATA) {
	ND_Unpacked_Free(U1);
	return 0;
    }
    /*
     *       .   
     *       .
     */
    if (l1 == -1 && rs_err != NO_DATA) {
ServFail:
	if (U1)
	    ND_Unpacked_Free(U1);
	dce = new_domain_entry();
	dce->dce_name = xbracketizedup(domain);
	dce->dce_cclass = 4;
	dce->dce_herror = TRY_AGAIN;
	dce->dce_time = time(NULL);
	dce->dce_expire = dce->dce_time + 600;	/* XXX fixed 10m */
	dce->dce_magic = DCE_MAGIC;
	*p = dce;
	return 1;
    }
    /*  NOERROR.  . */
    assert(U1);
    nseen = 0;
    for (UR = U1->Records; UR != NULL; UR = UR->NextRec) {
	if (UR->Section != 2 || UR->Type != T_A)
	    continue;
	++nseen;
	/* Anti sitefinder */
	if (UR->Data.Addr4 == 0x405E6E0B) {	/* 64.94.110.11 */
	    goto ReturnNxdomain;
	}
    }
    /*
     * nseen == 0   ,     A .
     *       .  - ,   
     *    NO_DATA.
     *    -  0 ("  ")
     */
    if (nseen == 0) {
	ND_Unpacked_Free(U1);
	return 0;
    }
    dce = new_domain_entry();
    *p = dce;
    dce->dce_name = xbracketizedup(domain);
    dce->dce_cclass = 2;
    dce->dce_magic = DCE_MAGIC;
    dce->dce_time = time(NULL);
    for (UR = U1->Records; UR != NULL; UR = UR->NextRec) {
	struct ipaddr_item *ia;

	if (UR->Section != 2 || UR->Type != T_A)
	    continue;
	ia = xcalloc(1, sizeof(*ia));
	ia->iat_addr = UR->Data.Addr4;
	ia->iat_hostname = xstrdup(domain);
	TAILQ_INSERT_TAIL(&dce->dce_ips, ia, iat_link);
	if (UR->Ttl < ttl)
	    ttl = UR->Ttl;
    }
    if (ttl > 86400) ttl = 86400;
    dce->dce_expire = dce->dce_time + ttl;
    ND_Unpacked_Free(U1);
    return 1;
}

static char *
xbracketizedup(const char *d)
{
    size_t l = strlen(d);
    char *r;

    r = xcalloc(1, l + 3);
    r[0] = '[';
    memmove(&r[1], d, l);
    r[l + 1] = ']';
    r[l + 2] = 0;
    return r;
}

static rescache_entry *
get_rescache_entry(void)
{
    int rc;
    rescache_entry *re;
    time_t now;

    if (0 != (rc = pthread_mutex_lock(&rescache_mx)))
	errc(1, rc, "pthread_mutex_lock[%d]", __LINE__);
    while ((re = TAILQ_FIRST(&rescache))) {
	TAILQ_REMOVE(&rescache, re, sr_link);
        time(&now);
	if (re->sr_expire <= now) {
            res_ndestroy(&re->sr_res);
	    free(re);
	} else break;	/* valid entry, so get out */
    }
    pthread_mutex_unlock(&rescache_mx);
    if (re)
	return re;
    time(&now);
    re = xcalloc(1, sizeof(*re));
    if (res_ninit(&re->sr_res) < 0)
	errx(1, "res_ninit()");
    re->sr_expire = now + 60;	/* 1 minute is normal timeout to reread */
    return re;
}

static void
put_rescache_entry(rescache_entry *re)
{
    int rc;

    if (0 != (rc = pthread_mutex_lock(&rescache_mx)))
	errc(1, rc, "pthread_mutex_lock[%d]", __LINE__);
    TAILQ_INSERT_TAIL(&rescache, re, sr_link);
    pthread_mutex_unlock(&rescache_mx);
}

static struct address_entry *
new_address_entry(void)
{
    struct address_entry *t;

    t = xcalloc(1, sizeof(*t));
    TAILQ_INIT(&t->ace_mmsg);
    return t;
}

static void
address_entry_mladd(struct address_entry * a, const char *format,...)
{
    char buf[1024];
    va_list ap;
    struct ml_entry *e;

    va_start(ap, format);
    vsnprintf(buf, sizeof buf, format, ap);
    va_end(ap);
    buf[sizeof buf - 1] = 0;
    e = xcalloc(1, sizeof(*e));
    e->mle_text = xstrdup(buf);
    TAILQ_INSERT_TAIL(&a->ace_mmsg, e, mle_link);
}

static struct address_entry *
address_entry_xclone(const struct address_entry * s)
{
    struct address_entry *t;
    struct ml_entry *mle;

    assert(s->ace_magic == ACE_MAGIC);
    t = xcalloc(1, sizeof(*t));
    bzero(t, sizeof(*t));
    t->ace_addr = xstrdup(s->ace_addr);
    t->ace_cclass = s->ace_cclass;
    t->ace_time = s->ace_time;
    t->ace_expire = s->ace_expire;
    t->ace_msg = s->ace_msg ? xstrdup(s->ace_msg) : NULL;
    TAILQ_INIT(&t->ace_mmsg);
    t->ace_magic = ACE_MAGIC;
    TAILQ_FOREACH(mle, &s->ace_mmsg, mle_link) {
	struct ml_entry *n = xcalloc(1, sizeof(*n));

	n->mle_text = xstrdup(mle->mle_text);
	TAILQ_INSERT_TAIL(&t->ace_mmsg, n, mle_link);
    }
    return t;
}

void
address_entry_free(struct address_entry *s)
{
    struct ml_entry *mle;

    if (!s)
	return;
    free(s->ace_msg);
    free(s->ace_addr);
    while ((mle = TAILQ_FIRST(&s->ace_mmsg))) {
	TAILQ_REMOVE(&s->ace_mmsg, mle, mle_link);
	free(mle->mle_text);
	free(mle);
    }
    free(s);
}

static struct domain_entry *
new_domain_entry(void)
{
    struct domain_entry *t;

    t = xcalloc(1, sizeof(*t));
    TAILQ_INIT(&t->dce_mxes);
    TAILQ_INIT(&t->dce_ips);
    return t;
}

static struct domain_entry *
domain_entry_xclone(const struct domain_entry *s)
{
    struct domain_entry *t;
    struct mx_item *mx;
    struct ipaddr_item *ia;

    assert(s);
    assert(s->dce_magic == DCE_MAGIC);
    t = xcalloc(1, sizeof(*t));
    bzero(t, sizeof(*t));
    assert(s->dce_name);
    t->dce_name = xstrdup(s->dce_name);
    t->dce_cclass = s->dce_cclass;
    t->dce_time = s->dce_time;
    t->dce_expire = s->dce_expire;
    t->dce_msg = s->dce_msg ? xstrdup(s->dce_msg) : NULL;
    TAILQ_INIT(&t->dce_mxes);
    TAILQ_FOREACH(mx, &s->dce_mxes, mxi_link) {
	struct mx_item *mm = xcalloc(1, sizeof(*mm));

	mm->mxi_name = xstrdup(mx->mxi_name);
	TAILQ_INSERT_TAIL(&t->dce_mxes, mm, mxi_link);
    }
    TAILQ_INIT(&t->dce_ips);
    TAILQ_FOREACH(ia, &s->dce_ips, iat_link) {
	struct ipaddr_item *in = xcalloc(1, sizeof(*in));

	in->iat_addr = ia->iat_addr;
	in->iat_hostname = xstrdup(ia->iat_hostname);
	TAILQ_INSERT_TAIL(&t->dce_ips, in, iat_link);
    }
    t->dce_magic = DCE_MAGIC;
    return t;
}

static void
domain_entry_free(struct domain_entry *s)
{
    struct mx_item *mx;
    struct ipaddr_item *i;

    if (!s)
	return;
    while ((i = TAILQ_FIRST(&s->dce_ips))) {
	TAILQ_REMOVE(&s->dce_ips, i, iat_link);
	free(i->iat_hostname);
	free(i);
    }
    while ((mx = TAILQ_FIRST(&s->dce_mxes))) {
	TAILQ_REMOVE(&s->dce_mxes, mx, mxi_link);
	free(mx->mxi_name);
	free(mx);
    }
    free(s->dce_msg);
    free(s->dce_name);
    free(s);
}

static void
free_ipaddrs(ipaddr_list *head)
{
    struct ipaddr_item *ipaddr;

    while ((ipaddr = TAILQ_FIRST(head))) {
	TAILQ_REMOVE(head, ipaddr, iat_link);
	free(ipaddr->iat_hostname);
	free(ipaddr);
    }
}

/* Convert string to small letters */

static void
lc(char **oldstr)
{
    char * p = *oldstr;

    while (*p) {
        *p = tolower(*p);
	p++;
    }
}

static void
extmap_read(void)
{
	FILE *fp;
	char buf[2000];
	char *key, *value;
	char pebuf[sizeof(AVL_IX_REC)+2000];
	AVL_IX_REC *pe = (AVL_IX_REC *) pebuf;

	if ((fp = fopen(extmap_path, "r"))) {
		avl_destroy_index(&extmap);
		avl_create_index(&extmap, AVL_NO_DUP_KEYS, 0);
		while (fgets(buf, sizeof(buf), fp)) {
			key = buf;
			key += strspn(key, " \t\r\n");
			if (*key == '#' || *key == ';') continue;
			value = key + strcspn(key, " \t\r\n");
			*value = 0;
			value++;
			value += strspn(value, " \t\r\n");
			lc(&key);
			strcpy(pe->key, key);
			switch(*value) {
				case '2':
					pe->recptr = (void *)&pass_code;
					break;
				case '4':
					pe->recptr = (void *)&tmpfail_code;
					break;
				case '5':
					pe->recptr = (void *)&reject_code;
					break;
				default:
					continue;
			}
			avl_add_key(pe, &extmap);
		}
		fclose(fp);
	}
	return;
}

int
check_extmap(char * address, unsigned long relay)
{
    char pebuf[sizeof(AVL_IX_REC)+2000];
    AVL_IX_REC *pe = (AVL_IX_REC *) pebuf;
    char *domain;
    size_t dl;
    time_t now;
    char ipbuf[16];
    struct in_addr ia;
    int rc;

    time(&now);
    lc(&address);

    if (extmap_path) {

	/* First, lookup relay ip address */

	ia.s_addr = htonl(relay);
	inet_ntop(AF_INET, &ia, ipbuf, sizeof ipbuf);
	sprintf(pe->key, "relay:%s", ipbuf);

	rc = pthread_mutex_lock(&extmap_mx);
	if (rc != 0)
	    errc(1, rc, "pthread_mutex_lock[%d]", __LINE__);

	if (difftime(now, extmap_opened) >= 60) {
	    extmap_read();
	    time(&extmap_opened);
	}

	if (avl_find_key(pe, &extmap) == AVL_IX_OK) {
	    pthread_mutex_unlock(&extmap_mx);
	    return *(int *)pe->recptr;
	}

	/* Second, lookup full recipient address */

	sprintf(pe->key, "to:%s", address);

	if (avl_find_key(pe, &extmap) == AVL_IX_OK) {
	    pthread_mutex_unlock(&extmap_mx);
	    return *(int *)pe->recptr;
	}

	/* Third, lookup recipient domain part */

	domain = strrchr(address, '@');
	if (!domain) {
	    pthread_mutex_unlock(&extmap_mx);
	    return -1;	/* XXX */
	}
	domain++;	/* now domain points to domain part of address */

	sprintf(pe->key, "todom:%s", domain);

	dl = strlen(pe->key);

	if (*(pe->key) && pe->key[dl-1] == '.') {
		dl--;
		pe->key[dl] = 0;
	}

	if (avl_find_key(pe, &extmap) == AVL_IX_OK) {
	    pthread_mutex_unlock(&extmap_mx);
	    return *(int *)pe->recptr;
	}
	pthread_mutex_unlock(&extmap_mx);
    }

    /* Well, we didn't find anything, so give up =) */

    return 0;
}
